/*
** PHREL
** $Id: db.c,v 1.5 2013/04/21 02:23:08 sella Exp $
** Copyright (c) 2004-2013 James M. Sella. All Rights Reserved.
** Released under the GPL Version 2 License.
** http://www.digitalgenesis.com
*/

static const char rcsid[] __attribute__((used)) = "$Id: db.c,v 1.5 2013/04/21 02:23:08 sella Exp $";

#include "db.h"
#include "prefs.h"

#include <string.h>
#include <syslog.h>
#include <assert.h>
#include <pthread.h>

/* Global Variables */
int connected = 0;
time_t last_connect = 0;

#ifdef HAVE_MYSQL
MYSQL mysql;

/* External Global Variables */
extern struct args_t args;
extern struct prefs_t prefs;

extern pthread_mutex_t db_mutex;

int db_library_init(void) {
#if defined(HAVE_LIBMYSQL_LIBRARY_INIT) || defined(mysql_library_init)
	if (mysql_library_init(0 , NULL, NULL)) {
#else
	if (mysql_server_init(0 , NULL, NULL)) {
#endif
		syslog(LOG_CRIT, "mysql_library_init(): failed to initialize MySQL library");
		return 1;
	}

	return 0;
}

void db_library_end(void) {
#if defined(HAVE_LIBMYSQL_LIBRARY_END) || defined(mysql_library_init)
	mysql_library_end();
#else
	mysql_server_end();
#endif
}

int db_connect(void) {
	unsigned int timeout = DEF_DBTIMO;
#ifdef MYSQL_OPT_RECONNECT
	unsigned int reconnect = 0;
#endif

	db_sanity();

	mysql_init(&mysql);
	/* Read my.cnf pulling in options from [phreld] section */
	mysql_options(&mysql, MYSQL_READ_DEFAULT_GROUP, PROGRAM_NAME);
	mysql_options(&mysql, MYSQL_OPT_CONNECT_TIMEOUT, (const char *) &timeout);
#ifdef MYSQL_OPT_RECONNECT
	mysql_options(&mysql, MYSQL_OPT_RECONNECT, (const char *) &reconnect);
#endif

	if (!mysql_real_connect(&mysql, args.db_host, args.db_user, args.db_pass, args.db_name, args.db_port, args.db_sock, args.db_comp)) {
		syslog(LOG_CRIT, "mysql_real_connect(%d): %s", mysql_errno(&mysql), mysql_error(&mysql));
		return 1;
	}

	connected = 1;
	last_connect = time(NULL);

	if (args.debug >= DEBUG_SQL) {
		syslog(LOG_DEBUG, "Connected to database");
	}

	return 0;
}

void db_disconnect(void) {
	db_sanity();

	if (connected == 1) {
		connected = 0;
		last_connect = 0;
		mysql_close(&mysql);

		if (args.debug >= DEBUG_SQL) {
			syslog(LOG_DEBUG, "Disconnected from database");
		}
	}
}

int db_check(void) {
	int error;

	db_sanity();

	if (connected != 1) { /* Skip if no connection */
		return 1;
	}

	if ((error = mysql_ping(&mysql)) != 0) {
		syslog(LOG_CRIT, "mysql_ping(%d): %s", mysql_errno(&mysql), mysql_error(&mysql));
	}

	return error;
}

int db_isconnected(void) {
	db_sanity();

	return connected;
}

int db_query(MYSQL_RES** res, char* sql) {
	int len;

	assert(res != NULL);
	assert(sql != NULL);

	db_sanity();

	len = strlen(sql);

	if (args.debug >= DEBUG_SQL) {
		if (args.test == 0) {
			syslog(LOG_DEBUG, "SQL: %s", sql);
		} else {
			syslog(LOG_DEBUG, "SQL: %s [test mode]", sql);
		}
	}

	if (args.test != 0) {
		return 0;
	}

	if (connected != 1) {
		if (args.debug >= DEBUG_INTERNAL) {
			syslog(LOG_WARNING, "skipping db_query() - no connection to database");
		}

		return 1;
	}

	if (mysql_real_query(&mysql, sql, len) != 0) {
		syslog(LOG_CRIT, "mysql_real_query(%d): %s", mysql_errno(&mysql), mysql_error(&mysql));

		return 1;
	}

	*res = mysql_store_result(&mysql);
	if (mysql_errno(&mysql) != 0) {
		syslog(LOG_CRIT, "mysql_store_result(%d): %s", mysql_errno(&mysql), mysql_error(&mysql));

		return 1;
	}

	return 0;
}

int db_sanity(void) {
	if (pthread_mutex_trylock(&db_mutex) == 0) {
		syslog(LOG_DEBUG, "DEBUG: db_mutex_sanity() failed - db_mutex was not locked on call to db function");
		
		pthread_mutex_unlock(&db_mutex);

		return 1;
	}

	return 0;
}

time_t db_last_connect(void) {
	return last_connect;
}
#endif

/*
** Local Variables:
** c-basic-offset: 3
** tab-width: 3
** End:
** vim: noet ts=3 sw=3
*/
